import scanpy as sc
import pandas as pd
import numpy as np

def supervised_spot_clust(idata, adata, label, portion=0.1, n_cluster=None, n_neighbors=100, min_dist=1):
    import umap
    idata.uns['cell_meta'][label+'_int'] = idata.uns['cell_meta'][label].astype('category').cat.codes
    masked_target = idata.uns['cell_meta'][label+'_int']
    np.random.seed(52)
    masked_target[np.random.choice(len(idata.uns['cell_meta'][label]), size=int(len(idata.uns['cell_meta'][label]) * portion), replace=False)] = -1
    embedding = umap.UMAP(n_neighbors=n_neighbors, min_dist=min_dist,random_state=52).fit_transform(idata.uns['cell_pattern'], y=masked_target)
    adata.obsm['X_umap'] = embedding
    sc.pp.neighbors(adata, n_neighbors=20, use_rep='X_umap')
    sc.tl.draw_graph(adata)
    flag = 1
    res_max = 10
    res_min = 0.0001
    res = 0.05
    if not n_cluster:
        k = len(idata.uns['cell_meta'][label].unique())
    else:
        k = n_cluster
    repeat = 0
    while (flag):
        sc.tl.leiden(adata, resolution=res)
        repeat+=1
        # print(len(adata.obs['leiden'].unique()), res)
        if (len(adata.obs['leiden'].unique()) == k) | (res < 0) | (res > 10) | (repeat > 1000):
            flag = 0
        elif len(adata.obs['leiden'].unique()) < k:
            res_min = res
            res = (res+res_max)/2
        else:
            res_max = res
            res = (res+res_min)/2
    adata.obs['leiden_supervised'] = adata.obs['leiden']
    
            
def unsupervised_spot_clust(idata, adata, n_cluster=None):
    import umap

    embedding = umap.UMAP(n_neighbors=100, min_dist=1,random_state=52).fit_transform(idata.uns['cell_pattern'])
    adata.obsm['X_umap'] = embedding
    sc.pp.neighbors(adata, n_neighbors=20, use_rep='X_umap')
    sc.tl.draw_graph(adata)
    flag = 1
    res_max = 10
    res_min = 0.0001
    res = 0.05
    k = n_cluster
    while (flag):
        sc.tl.leiden(adata, resolution=res)
        # print(len(adata.obs['leiden'].unique()), res)
        if (len(adata.obs['leiden'].unique()) == k) | (res < 0) | (res > 10):
            flag = 0
        elif len(adata.obs['leiden'].unique()) < k:
            res_min = res
            res = (res+res_max)/2
        else:
            res_max = res
            res = (res+res_min)/2
    adata.obs['leiden_unsupervised'] = adata.obs['leiden']

def interface_clustering(idata, n_cluster=15):
    flag = 1
    res_max = 10
    res_min = 0.0001
    res = 0.05
    k = n_cluster
    sc.pp.neighbors(idata, n_neighbors=52, use_rep='X_umap')
    while (flag):
        sc.tl.leiden(idata, resolution=res,random_state=52)
        # print(len(idata.obs['leiden'].unique()), res)
        if (len(idata.obs['leiden'].unique()) == k) | (res < 0) | (res > 10):
            flag = 0
        elif len(idata.obs['leiden'].unique()) < k:
            res_min = res
            res = (res+res_max)/2
        else:
            res_max = res
            res = (res+res_min)/2


import scanpy as sc
import numpy as np
import matplotlib.pyplot as plt
import squidpy as sq

def paga(idata, label):
    import umap, umap.plot
    embedding = umap.UMAP(n_neighbors=30, min_dist=1,random_state=52).fit_transform(idata.obsm['smooth_pattern_score'])
    idata.obsm['X_umap'] = embedding
    sc.pl.umap(idata,color='label')
    sc.pp.neighbors(idata, n_neighbors=20, use_rep='X_umap')
    sc.tl.paga(idata, groups=label)
    
def pseudotime(idata, root_label):
    idata.uns['iroot'] = np.flatnonzero(idata.obs['label']  == root_label)[0]
    sc.tl.dpt(idata)
    
def projection(idata):
    import cellrank as cr
    sq.gr.spatial_neighbors(idata, key_added='spatial')
    pt=cr.kernels.PseudotimeKernel(idata, time_key='dpt_pseudotime')
    pt.compute_transition_matrix()
    ck = cr.kernels.ConnectivityKernel(idata,conn_key='spatial_connectivities')
    ck.compute_transition_matrix()
    k=8*pt+2*ck
    k.compute_transition_matrix()
    k.plot_projection(basis="spatial")
    



import pandas as pd


from . import svi
from . import preprocess
from . import clustering
from . import enrichment
from . import visualization
from . import util
from . import trajectory

class SPIDER():
    def __init__(self):
        self.svi = svi
        self.pp = preprocess
        self.cl = clustering
        self.er = enrichment
        self.vis = visualization
        self.util = util
        self.traj = trajectory
        pass

    def prep(self,
            adata_input, work_dir, R_path,
            no_spatalk=False,
            cluster_key='type', 
            is_human=True, 
            n_neighs=6, 
            coord_type='generic', 
            imputation=True,
            overwrite=False,
    ):
        adata = adata_input.copy()
        del adata_input
        # detect cci

        # Step 1
        interface_cell_pair, interface_meta = preprocess.find_interfaces(adata, coord_type=coord_type, n_neighs=n_neighs, cluster_key=cluster_key)
        # Step 2
        lr_raw = preprocess.subset_lr(adata, no_spatalk, work_dir, cluster_key, is_human, overwrite, R_path)
        lr_df, adata = preprocess.subset_adata(adata, lr_raw)
        score = preprocess.score(adata, lr_df, interface_cell_pair, imputation)
        # Idata object construction
        idata = preprocess.idata_construct(score, interface_meta, lr_df, lr_raw, adata)
        return idata

    def find_svi(self, idata, out_f, R_path, abstract=True, overwrite=False, n_neighbors=10, threshold=0.01, pattern_prune_threshold=0.0001):
        from os.path import exists
        from os import mkdir
        if not exists(out_f):
            print(f'Creating folder {out_f}')
            mkdir(out_f)
        if len(idata) < 200:
            print('number of interface is less than 200, skipping abstraction')
            abstract=False
        if abstract:
            som, idata, meta_idata = svi.abstract(idata, n_neighbors)
            svi.find_svi(meta_idata,out_f, overwrite, R_path, som=som) #generating results
            print('finished running all SVI tests')
            svi_df, svi_df_strict = svi.combine_SVI(meta_idata,threshold=threshold)
            if (overwrite) | (not exists(f'{out_f}pattern.csv')):
                svi.SVI_patterns(meta_idata, svi_df_strict, pattern_prune_threshold=pattern_prune_threshold)
                pd.DataFrame(meta_idata.obsm['pattern_score']).to_csv(f'{out_f}pattern.csv')
                meta_idata.var.to_csv(f'{out_f}membership.csv')
            else:
                meta_idata.obsm['pattern_score'] = pd.read_csv(f'{out_f}pattern.csv', index_col=0).to_numpy()
                meta_idata.var = pd.read_csv(f'{out_f}membership.csv', index_col=0)
            svi.meta_pattern_to_idata(idata, meta_idata)
            pd.DataFrame(meta_idata.obsm['pattern_score']).to_csv(f'{out_f}full_pattern.csv')
        else:
            svi.find_svi(idata, out_f, R_path, overwrite) #generating results
            svi_df, svi_df_strict = svi.combine_SVI(idata,threshold=threshold)
            if (overwrite) | (not exists(f'{out_f}pattern.csv')):
                svi.SVI_patterns(idata, svi_df_strict, pattern_prune_threshold=pattern_prune_threshold)
                pd.DataFrame(idata.obsm['pattern_score']).to_csv(f'{out_f}pattern.csv')
                idata.var.to_csv(f'{out_f}membership.csv')
            else:
                idata.obsm['pattern_score'] = pd.read_csv(f'{out_f}pattern.csv', index_col=0)
                idata.var = pd.read_csv(f'{out_f}membership.csv', index_col=0)   
            meta_idata = None
        idata.var[[f'pattern_correlation_{x}' for x in range(idata.obsm['pattern_score'].shape[1])]] = 0
        corr_df=pd.concat([idata[:,idata.var['is_svi']==1].to_df(),pd.DataFrame(idata.obsm['pattern_score'],index=idata.obs_names)],axis=1).corr().loc[idata[:,idata.var['is_svi']==1].var_names, range(idata.obsm['pattern_score'].shape[1])]
        idata.var.loc[idata[:,idata.var['is_svi']==1].var_names, [f'pattern_correlation_{x}' for x in range(idata.obsm['pattern_score'].shape[1])]] = corr_df.to_numpy()
        return idata, meta_idata
        
    def cell_transform(self, idata, adata, label=None):
        from scanpy.tools import rank_genes_groups
        import anndata
        adata = adata[adata.obs_names.isin(idata.uns['cell_meta'].index)]
        util.scored_spot_interface(idata)
        util.interaction_spot_interface(idata)
        adata.obsm['interaction_pattern'] = idata.uns['cell_pattern'].loc[adata.obs_names]
        adata.obsm['interaction_score'] = idata.uns['cell_score'].loc[adata.obs_names]
        print(f'Added key interaction_pattern, interaction_score in adata.obsm')
        
        if label is not None:
            adata_lri = anndata.AnnData(idata.uns['cell_score'])
            idata.uns['cell_meta'][label] = idata.uns['cell_meta'][label].astype(str).astype('category')
            small_clust = idata.uns['cell_meta'][label].value_counts()[idata.uns['cell_meta'][label].value_counts()==1].index.to_numpy()
            adata_lri.obs = idata.uns['cell_meta']
            adata_lri = adata_lri[~adata_lri.obs[label].isin(small_clust),:]
            rank_genes_groups(adata_lri, groupby=label)
            adata.uns['rank_interaction_score_groups'] = adata_lri.uns['rank_genes_groups']
            adata_pattern = anndata.AnnData(idata.uns['cell_pattern'])
            adata_pattern.obs = idata.uns['cell_meta']
            adata_pattern = adata_pattern[~adata_pattern.obs[label].isin(small_clust),:]
            rank_genes_groups(adata_pattern, groupby=label)
            adata.uns['rank_interaction_pattern_groups'] = adata_pattern.uns['rank_genes_groups']                                                          
            print(f'Added key rank_interaction_score_groups, rank_interaction_pattern_groups in adata.uns')   
        adata.obsm['interaction_pattern'] = adata.obsm['interaction_pattern'].to_numpy()                                                   
        adata.obsm['interaction_score'] = adata.obsm['interaction_score'].to_numpy()                                                   
        return adata, adata_lri, adata_pattern
    







import pandas as pd
import numpy as np
import scanpy as sc
import networkx as nx
import magic
import scprep
from pathlib import Path
from sklearn.metrics import pairwise_distances
from scipy.spatial import Delaunay
from sklearn.decomposition import PCA,NMF
import anndata 
from importlib import resources


# detect cci
def cci_spatalk(adata, work_dir, cluster_key, is_human, out_f, R_path):
    import os
    count_f = f'{work_dir}/adata_count.csv'
    meta_f = f'{work_dir}/adata_meta.csv'
    df = adata.to_df()
    df.index = "C"+df.index
    df.to_csv(count_f)
    meta = adata.obs[cluster_key].reset_index()
    meta[['x', 'y']] = adata.obsm['spatial']
    meta.columns = ['cell','celltype', 'x', 'y']
    meta.cell = "C"+meta.cell
    meta = meta[['cell','x', 'y', 'celltype']]
    if not pd.api.types.is_string_dtype(meta.celltype.dtype):
        meta.celltype = "T"+meta.celltype.astype('str')
    meta.celltype = meta.celltype.str.replace(' ', '_')
    meta.celltype = meta.celltype.str.replace('-', '_')
    meta.to_csv(meta_f)
    species = 'Human' if is_human else 'Mouse'
    with resources.path("spider.R_script", "run_spatalk.R") as pw_fn:
        os.system(str(f'/bin/bash -c "{R_path} -f {pw_fn} {count_f} {meta_f} {species} {out_f}"'))
        # os.system(str(f'/bin/bash -c "source /etc/profile;module load GCC/11.2.0 OpenMPI/4.1.1 R/4.2.0 Anaconda3/2022.05 R-bundle-Bioconductor/3.15-R-4.2.0;R -f {pw_fn} {count_f} {meta_f} {species} {out_f}"'))
    
# imputation
def impute_MAGIC(adata):
    magic_op = magic.MAGIC(n_jobs=5)
    inp = adata.to_df()
    inp = scprep.normalize.library_size_normalize(inp)
    inp = scprep.transform.sqrt(inp)
    outp = magic_op.fit_transform(inp)
    adata.X = outp

# idata
def idata_construct(score, pairs_meta, lr_df, lr_raw, adata):
    idata = anndata.AnnData(score)
    idata.obs_names = pairs_meta.index
    idata.var_names = lr_df.index
    idata.uns['lr_meta'] = lr_raw
    idata.obs = pairs_meta
    unique_cells = np.unique(idata.obs[['A', 'B']].to_numpy().flatten())
    cell_meta = adata.obs.loc[unique_cells]
    idata.uns['cell_meta'] = cell_meta
    # quality check
    sc.pp.calculate_qc_metrics(idata, inplace=True, percent_top=None)
    sc.pp.filter_genes(idata, min_cells=5)
    sc.pp.filter_cells(idata, min_genes=1)
    idata.obsm['spatial'] = idata.obs[['row', 'col']].to_numpy()
    print(f'Construct idata with {idata.shape[0]} interfaces and {idata.shape[1]} LR pairs.')
    return idata

def subset_lr(adata, no_spatalk, work_dir, cluster_key, is_human, overwrite, R_path):
    from os.path import exists
    if not no_spatalk:
        out_f = f'{work_dir}/spatalk'
        if overwrite | (not exists(f'{out_f}_lrpair.csv')):
            cci_spatalk(adata, work_dir, cluster_key, is_human, out_f, R_path)
        else:
            print(f'{out_f}_lrpair.csv already exists, skipping spatalk.')
    try:
        print('using spatalk result')
        lr_raw = pd.read_csv(f'{out_f}_lrpair.csv', index_col=0).sort_values('score')
        lr_raw = lr_raw.drop_duplicates(subset=['ligand', 'receptor'], keep="last")
    except:
        print('no spatalk result, using all lrpairs')
        lr_raw = load_lr_df(is_human).drop_duplicates(subset=['ligand', 'receptor'], keep="last")
        lr_raw['score'] = 1
    return lr_raw


def score(adata, lr_df, pairs, imputation):
    if imputation:
        print('Running imputation with MAGIC')
        impute_MAGIC(adata)
    
    exp_ref = adata.to_df()
    exp_ref = exp_ref.loc[:,~exp_ref.columns.duplicated()]
    l = lr_df['ligand'].to_numpy().flatten()
    r = lr_df['receptor'].to_numpy().flatten()
    sub_exp = exp_ref[np.concatenate((l, r))].to_numpy()
    sub_exp_rev = exp_ref[np.concatenate((r, l))].to_numpy()
    edge_exp_both = np.multiply(sub_exp[pairs[0]], sub_exp_rev[pairs[1]])
    # equation 2 in the manuscript
    print('scoring')
    score = lr_df['score'].to_numpy()*np.sqrt(np.maximum(edge_exp_both[:, :int(len(l))], edge_exp_both[:, int(len(l)):]))
    return score

def find_interfaces(adata, coord_type, n_neighs, cluster_key):
        pairs = find_pairs(adata, coord_type=coord_type, n_neighs=n_neighs)
        pairs_meta = meta(adata, cluster_key, pairs)
        return pairs, pairs_meta

def subset_adata(adata, lr_df):
    genes = adata.var_names.tolist()
    lr_df = lr_df[lr_df['ligand'].isin(genes) & lr_df['receptor'].isin(genes)]
    lr_df.index = lr_df['ligand'] + "_" + lr_df['receptor']
    l = lr_df['ligand'].to_numpy().flatten()
    r = lr_df['receptor'].to_numpy().flatten()
    unique_lr = np.unique(np.concatenate((l, r)))
    adata = adata[:, adata.var_names.isin(unique_lr)]
    sc.pp.filter_genes(adata, min_cells=1)
    sc.pp.normalize_total(adata, target_sum=1e4)
    genes = adata.var_names.tolist()
    lr_df = lr_df[lr_df['ligand'].isin(genes) & lr_df['receptor'].isin(genes)]
    return lr_df, adata

def find_pairs(adata, coord_type='generic', n_neighs=6):
    from sklearn.metrics import pairwise_distances
    from scipy.sparse import triu
    from scipy.spatial import Delaunay
    from squidpy.gr import spatial_neighbors
    if coord_type=='grid':    
        spatial_neighbors(adata, coord_type=coord_type, n_neighs=n_neighs)
        return np.transpose(triu(adata.obsp['spatial_connectivities']).nonzero()).T
    else:
        print('non grid coordinate')
        spatial_neighbors(adata, coord_type=coord_type, delaunay=True, n_neighs=n_neighs)
        potential_pairs = np.transpose(triu(adata.obsp['spatial_connectivities']).nonzero()).T
        print('distance cutoff')
        d = np.linalg.norm(adata.obsm['spatial'][potential_pairs[0]]-adata.obsm['spatial'][potential_pairs[1]], axis=1)
        d = (d - d.mean()) / d.std()**2
        potential_pairs = potential_pairs[:, d <= d.std()]
        return potential_pairs

def meta(adata, cluster_key, pairs):
    # get label
    pairs_meta = pd.DataFrame()
    pairs_meta['A'] = adata.obs_names[pairs[0]]
    pairs_meta['B'] = adata.obs_names[pairs[1]]
    pairs_meta[['A_row', 'A_col']] = adata.obsm['spatial'][pairs[0]]
    pairs_meta[['B_row', 'B_col']] =  adata.obsm['spatial'][pairs[1]]

    if cluster_key != '': 
        node_labels_text = adata.obs[cluster_key].to_numpy()
        pairs_meta['A_label'] = node_labels_text[pairs[0]].astype(str)
        pairs_meta['B_label'] = node_labels_text[pairs[1]].astype(str)
        node_labels = adata.obs[cluster_key].astype('category').cat.codes.to_numpy()
        pairs_meta['A_label_int'] = node_labels[pairs[0]]
        pairs_meta['B_label_int'] = node_labels[pairs[1]]
        pairs_meta['label_1'] = pairs_meta["A_label_int"].astype(str) + pairs_meta["B_label_int"].astype(str)
        pairs_meta['label_2'] = pairs_meta["B_label_int"].astype(str) + pairs_meta["A_label_int"].astype(str)
        pairs_meta['label_int'] = pairs_meta[['label_1', 'label_2']].astype(int).max(axis=1).astype(str).astype('category')
        label_1 = pairs_meta['A_label'].astype(str) + '_' + pairs_meta['B_label'].astype(str).to_numpy()
        label_2 = pairs_meta['B_label'].astype(str) + '_' + pairs_meta['A_label'].astype(str).to_numpy()
        pick = pairs_meta[['label_1', 'label_2']].astype(int).idxmax(axis=1).to_numpy()
        text_label = [label_1[i] if x=='label_1' else label_2[i] for i,x in enumerate(pick)]
        pairs_meta['label'] = text_label
        pairs_meta['label'] = pairs_meta['label'].astype('category')

    pairs_meta.index = pairs_meta['A'] + "_" + pairs_meta['B']

    # get position  
    A_pos = pairs_meta[['A_row', 'A_col']].to_numpy(dtype=float)
    B_pos = pairs_meta[['B_row', 'B_col']].to_numpy(dtype=float)
    # optimal solution for Equation 1 in the manuscript
    avg_pair_pos = (A_pos + B_pos) / 2
    pairs_meta[['row', 'col']] = avg_pair_pos
    pairs_meta['dist'] = np.linalg.norm(A_pos-B_pos, axis=1)
    return pairs_meta


def load_lr_df(is_human):
    from importlib import resources
    with resources.path("spider.lrdb", "lrpairs.tsv") as pw_fn:
        lr_list = pd.read_csv(pw_fn, sep='\t', index_col=0)
    if is_human:
        print('Using human LR pair dataset.')
        lr_list = lr_list[lr_list.species=='Human']
    else:
        print('Using mouse LR pair dataset.')
        lr_list = lr_list[lr_list.species=='Mouse']
    return lr_list






import pandas as pd

def relabel_interface(idata, clust_key):
    node_labels_text = idata.uns['cell_meta'][clust_key]
    idata.obs[f'A_label_{clust_key}'] = node_labels_text.loc[idata.obs['A']].astype(str).to_numpy()
    idata.obs[f'B_label_{clust_key}'] = node_labels_text.loc[idata.obs['B']].astype(str).to_numpy()
    node_labels = idata.uns['cell_meta'][clust_key].astype('category').cat.codes
    idata.obs[f'A_label_int_{clust_key}'] = node_labels.loc[idata.obs['A']].to_numpy()
    idata.obs[f'B_label_int_{clust_key}'] = node_labels.loc[idata.obs['B']].to_numpy()
    idata.obs['label_1'] = idata.obs[f'A_label_int_{clust_key}'].astype(str) + idata.obs[f'B_label_int_{clust_key}'].astype(str)
    idata.obs['label_2'] = idata.obs[f'B_label_int_{clust_key}'].astype(str) + idata.obs[f'A_label_int_{clust_key}'].astype(str)
    idata.obs[f'label_{clust_key}_int'] = idata.obs[['label_1', 'label_2']].astype(int).max(axis=1).astype(str).astype('category')
    label_1 = idata.obs[f'A_label_{clust_key}'].astype(str) + '_' + idata.obs[f'B_label_{clust_key}'].astype(str).to_numpy()
    label_2 = idata.obs[f'B_label_{clust_key}'].astype(str) + '_' + idata.obs[f'A_label_{clust_key}'].astype(str).to_numpy()
    pick = idata.obs[['label_1', 'label_2']].astype(int).idxmax(axis=1).to_numpy()
    text_label = [label_1[i] if x=='label_1' else label_2[i] for i,x in enumerate(pick)]
    idata.obs[f'label_{clust_key}'] = text_label
    idata.obs[f'label_{clust_key}'] = idata.obs[f'label_{clust_key}'].astype('category')
    print(f'Added key label_{clust_key} in idata.obs')
    
def scored_spot_interface(idata):       
    belonging = {}
    cells = idata.uns['cell_meta'].index
    for i in cells:
        belonging[i] = []
    for pair in idata.obs.reset_index()[['index','A', 'B']].to_numpy():
        belonging[pair[1]].append(pair[0])
        belonging[pair[2]].append(pair[0])
    score = pd.DataFrame(idata.obsm['pattern_score'], index=idata.obs_names)
    df = pd.concat([score.loc[belonging[c]].mean() for c in cells], axis=1).T     
    df.index = cells
    idata.uns['cell_pattern'] = df
    print(f'Added key cell_pattern in idata.uns')

def interaction_spot_interface(idata):       
    belonging = {}
    cells = idata.uns['cell_meta'].index
    for i in cells:
        belonging[i] = []
    for pair in idata.obs.reset_index()[['index','A', 'B']].to_numpy():
        belonging[pair[1]].append(pair[0])
        belonging[pair[2]].append(pair[0])
    score = idata.to_df()
    df = pd.concat([score.loc[belonging[c]].mean() for c in cells], axis=1).T     
    df.index = cells
    idata.uns['cell_score'] = df
    print(f'Added key cell_score in idata.uns')

            


import logging
from time import time
import warnings
from itertools import zip_longest
from typing import Optional, Dict, Tuple, Union, List

import numpy as np
import pandas as pd
from gpflow import default_float

from anndata import AnnData

from tqdm.auto import tqdm

from ._internal.util import DistanceCache
from ._internal.util import bh_adjust, calc_sizefactors, default_kernel_space, kspace_walk
from ._internal.score_test import (
    NegativeBinomialScoreTest,
    combine_pvalues,
)
from ._internal.tf_dataset import AnnDataDataset


def _add_individual_score_test_result(resultdict, kernel, kname, gene):
    if "kernel" not in resultdict:
        resultdict["kernel"] = [kname]
    else:
        resultdict["kernel"].append(kname)
    if "gene" not in resultdict:
        resultdict["gene"] = [gene]
    else:
        resultdict["gene"].append(gene)
    for key, var in vars(kernel).items():
        if key[0] != "_":
            if key not in resultdict:
                resultdict[key] = [var]
            else:
                resultdict[key].append(var)
    return resultdict


def _merge_individual_results(individual_results):
    merged = {}
    for res in individual_results:
        for k, v in res.items():
            if k not in merged:
                merged[k] = v if not np.isscalar(v) else [v]
            else:
                if isinstance(merged[k], np.ndarray):
                    merged[k] = np.concatenate((merged[k], v))
                elif isinstance(v, list):
                    merged[k].extend(v)
                else:
                    merged[k].append(v)
    return pd.DataFrame(merged)


def test(
    adata: AnnData,
    layer: Optional[str] = None,
    omnibus: bool = False,
    spatial_key: str = "spatial",
    kernel_space: Optional[Dict[str, Union[float, List[float]]]] = None,
    sizefactors: Optional[np.ndarray] = None,
    stack_kernels: Optional[bool] = None,
    use_cache: bool = True,
) -> Tuple[pd.DataFrame, Union[pd.DataFrame, None]]:
    """
    Test for spatially variable genes.

    Perform a score test to detect spatially variable genes in a spatial transcriptomics
    dataset. Multiple kernels can be tested to detect genes with different spatial patterns and lengthscales.
    The test uses a count-based likelihood and thus operates on raw count data. Two ways of handling multiple
    kernels are implemented: omnibus and Cauchy combination. The Cauchy combination tests each kernel separately
    and combines the p-values afterwards, while the omnibus test tests all kernels simultaneously. With multiple
    kernels the omnibus test is faster, but may have slightly less statistical power than the Cauchy combination.

    Args:
        adata: The annotated data matrix.
        layer: Name of the AnnData object layer to use. By default ``adata.X`` is used.
        omnibus: Whether to do an omnibus test.
        spatial_key: Key in ``adata.obsm`` where the spatial coordinates are stored.
        kernel_space: Kernels to test against. Dictionary with the name of the kernel function as key and list of
            lengthscales (if applicable) as values. Currently, three kernel functions are known:

            * ``SE``, the squared exponential kernel :math:`k(\\boldsymbol{x}^{(1)}, \\boldsymbol{x}^{(2)}; l) = \\exp\\left(-\\frac{\\lVert \\boldsymbol{x}^{(1)} - \\boldsymbol{x}^{(2)} \\rVert}{l^2}\\right)`
            * ``PER``, the periodic kernel :math:`k(\\boldsymbol{x}^{(1)}, \\boldsymbol{x}^{(2)}; l) = \\cos\\left(2 \pi \\frac{\\sum_i (x^{(1)}_i - x^{(2)}_i)}{l}\\right)`
            * ``linear``, the linear kernel :math:`k(\\boldsymbol{x}^{(1)}, \\boldsymbol{x}^{(2)}) = (\\boldsymbol{x}^{(1)})^\\top \\boldsymbol{x}^{(2)}`

            By default, 5 squared exponential and 5 periodic kernels with lengthscales spanning the range of the
            data will be used.
        sizefactors: Scaling factors for the observations. Default to total read counts.
        stack_kernels: When using the Cauchy combination, all tests can be performed in one operation by stacking
            the kernel matrices. This leads to increased memory consumption, but will drastically improve runtime
            on GPUs for smaller data sets. Defaults to ``True`` for datasets with less than 2000 observations and
            ``False`` otherwise.
        use_cache: Whether to use a pre-computed distance matrix for all kernels instead of computing the distance
            matrix anew for each kernel. Increases memory consumption, but is somewhat faster.

    Returns:
        If ``omnibus==True``, a tuple with a Pandas DataFrame as the first element and ``None`` as the second.
        The DataFrame contains the results of the test for each gene, in particular p-values and BH-adjusted p-values.
        Otherwise, a tuple of two DataFrames. The first contains the combined results, while the second contains results
        from individual tests.
    """
    logging.info("Performing DE test")

    X = adata.obsm[spatial_key]
    dcache = DistanceCache(X, use_cache)
    if sizefactors is None:
        sizefactors = calc_sizefactors(adata)
    if kernel_space is None:
        kernel_space = default_kernel_space(dcache)

    individual_results = None if omnibus else []
    if stack_kernels is None and adata.n_obs <= 2000 or stack_kernels or omnibus:
        kernels = []
        kernelnames = []
        for k, name in kspace_walk(kernel_space, dcache):
            kernels.append(k)
            kernelnames.append(name)
        test = NegativeBinomialScoreTest(
            sizefactors,
            omnibus,
            kernels,
        )

        results = []
        with tqdm(total=adata.n_vars) as pbar:
            for i, (y, g) in AnnDataDataset(adata, dtype=default_float(), layer=layer).enumerate():
                i = i.numpy()
                g = g.numpy().decode("utf-8")
                t0 = time()
                result, _ = test(y)
                t = time() - t0
                pbar.update()
                res = {"gene": g, "time": t}
                resultdict = result.to_dict()
                if omnibus:
                    res.update(resultdict)
                else:
                    res["pval"] = combine_pvalues(result).numpy()
                results.append(res)
                if not omnibus:
                    for k, n in zip(kernels, kernelnames):
                        _add_individual_score_test_result(resultdict, k, n, g)
                    individual_results.append(resultdict)
    else:  # doing all tests at once with stacked kernels leads to excessive memory consumption
        results = [[0, []] for _ in range(adata.n_vars)]
        nullmodels = []
        test = NegativeBinomialScoreTest(sizefactors)
        for k, n in kspace_walk(kernel_space, dcache):
            test.kernel = k
            if len(nullmodels) > 0:
                nullit = nullmodels
                havenull = True
            else:
                nullit = ()
                havenull = False
            with tqdm(total=adata.n_vars) as pbar:
                for null, (i, (y, g)) in zip_longest(
                    nullit, AnnDataDataset(adata, dtype=default_float(), layer=layer).enumerate()
                ):
                    i = i.numpy()
                    g = g.numpy().decode("utf-8")

                    t0 = time()
                    res, null = test(y, null)
                    t = time() - t0
                    if not havenull:
                        nullmodels.append(null)
                    pbar.update()
                    results[i][0] += t
                    results[i][1].append(res)
                    resultdict = res.to_dict()
                    individual_results.append(
                        _add_individual_score_test_result(resultdict, k, n, g)
                    )
        for i, g in enumerate(adata.var_names):
            results[i] = {
                "gene": g,
                "time": results[i][0],
                "pval": combine_pvalues(results[i][1]).numpy(),
            }

    results = pd.DataFrame(results)
    results["padj"] = bh_adjust(results.pval.to_numpy())

    if individual_results is not None:
        individual_results = _merge_individual_results(individual_results)
    return results, individual_results



from typing import Optional, List, Union, Tuple
from dataclasses import dataclass
from enum import Enum, auto
import logging
import warnings

import numpy as np
import pandas as pd
import tensorflow as tf
from gpflow.utilities.ops import square_distance

from anndata import AnnData

from ._internal.util import calc_sizefactors, dense_slice
from ._internal.util_mixture import prune_components, prune_labels


@dataclass(frozen=True)
class TissueSegmentationParameters:
    """
    Parameters for tissue segmentation.

    Args:
        nclasses: Maximum number of regions to consider. Defaults to the square root of the number of observations.
        neighbors: Number of neighbors for the nearest-neighbor graph. Defaults to a fully connected graph (there is
            no speed difference). A value of 0 makes the model ignore spatial information and reduces it to a Poisson
            mixture model with a Dirichlet process prior.
        smoothness_factor: Spatial smoothness. Larger values induce more fine-grained segmentations. This value will
            be multiplied with the minimum squared distance within the data set, so it is dimensionless. Defaults to ``2``.
        class_prune_threshold: Probability threshold at which unused regions are removed. Defaults to ``1e-6``.
        abstol: Absolute convergence tolerance. Defaults to ``1e-12``.
        reltol: Relative convergence tolerance. Defaults to ``1e-12``.
        maxiter: Maximum number of iterations. Defaults to ``1000``.
        gamma_1: Parameter of the Poisson mean prior, defaults to ``1e-14``.
        gamma_2: Parameter of the Poisson mean prior, defaults to ``1e-14``.
        eta_1: Parameter of the Dirichlet process hyperprior, defaults to ``1``.
        eta_2: Parameter of the Dirichlet process hyperprior, defaults to ``1``.
    """

    nclasses: Optional[int] = None
    neighbors: Optional[int] = None
    smoothness_factor: float = 2
    class_prune_threshold: float = 1e-6
    abstol: float = 1e-12
    reltol: float = 1e-12
    maxiter: int = 1000
    gamma_1: float = 1e-14
    gamma_2: float = 1e-14
    eta_1: float = 1
    eta_2: float = 1

    def __post_init__(self):
        assert (
            self.nclasses is None or self.nclasses >= 1
        ), "Number of classes must be None or at least 1"
        assert (
            self.neighbors is None or self.neighbors >= 0
        ), "Number of neighbors must be None or at least 0"
        assert self.smoothness_factor > 0, "Smoothness factor must be greater than 0"
        assert (
            self.class_prune_threshold >= 0 and self.class_prune_threshold <= 1
        ), "Class pruning threshold must be between 0 and 1"
        assert self.abstol > 0, "Absolute tolerance must be greater than 0"
        assert self.reltol > 0, "Relative tolerance must be greater than 0"
        assert self.maxiter >= 1, "Maximum number of iterations must be greater than or equal to 1"
        assert self.gamma_1 > 0, "Gamma1 hyperparameter must be greater than 0"
        assert self.gamma_2 > 0, "Gamma2 hyperparameter must be greater than 0"
        assert self.eta_1 > 0, "Eta1 hyperparameter must be greater than 0"
        assert self.eta_2 > 0, "Eta2 hyperparameter must be greater than 0"


class TissueSegmentationStatus(Enum):
    AbsoluteToleranceReached = auto()
    RelativeToleranceReached = auto()
    MaximumIterationsReached = auto()


@dataclass(frozen=True)
class TissueSegmentation:
    """
    Results of tissue segmentation.

    Args:
        converged: Whether the optimization converged.
        status: Status of the optimization.
        labels: The estimated region labels.
        class_probabilities: N_obs x N_regions array with the estimated region probabilities for each observation.
        gammahat_1: N_classes x N_genes array with the estimated parameter of the gene expression posterior.
        gammahat_2: N_classes x 1 array with the estimated parameter of the gene expression posterior.
        niter: Number of iterations for the optimization.
        prune_iterations: Iterations at which unneeded regions were removed.
        elbo_trace: ELBO values at each iteration.
        nclasses_trace: Number of regions at each iteration.
    """

    converged: bool
    status: TissueSegmentationStatus
    labels: np.ndarray
    class_probabilities: np.ndarray
    gammahat_1: np.ndarray
    gammahat_2: np.ndarray
    niter: int
    prune_iterations: np.ndarray
    elbo_trace: np.ndarray
    nclasses_trace: np.ndarray


@tf.function(experimental_relax_shapes=True)
def _segment(
    counts: tf.Tensor,
    sizefactors: tf.Tensor,
    distances: tf.Tensor,
    nclasses: tf.Tensor,
    fnclasses: tf.Tensor,
    ngenes: tf.Tensor,
    labels: tf.Tensor,
    gamma_1: tf.Tensor,
    gamma_2: tf.Tensor,
    eta_1: tf.Tensor,
    eta_2: tf.Tensor,
    alphahat_1: tf.Tensor,
    alphahat_2: tf.Tensor,
    etahat_2: tf.Tensor,
    gammahat_1: tf.Tensor,
    gammahat_2: tf.Tensor,
):
    eta_1_nclasses = eta_1 + fnclasses
    if labels is not None and distances is not None:
        p_x_neighborhood = tf.TensorArray(counts.dtype, size=tf.shape(gammahat_1)[0])
        for c in tf.range(tf.shape(gammahat_1)[0]):
            p_x_neighborhood = p_x_neighborhood.write(
                c, -tf.reduce_sum(tf.where(labels != c, distances, 0), axis=1)
            )
        p_x_neighborhood = p_x_neighborhood.stack()
        p_x_neighborhood = p_x_neighborhood - tf.reduce_logsumexp(
            p_x_neighborhood, axis=0, keepdims=True
        )
    else:
        p_x_neighborhood = tf.convert_to_tensor(0, counts.dtype)

    lambdahat_1 = gammahat_1 / gammahat_2
    lambdahat_2 = tf.math.digamma(gammahat_1) - tf.math.log(gammahat_2)
    alpha12 = alphahat_1 + alphahat_2
    dgalpha = tf.math.digamma(alpha12)
    vhat2 = tf.math.digamma(alphahat_1) - dgalpha
    vhat3 = tf.math.digamma(alphahat_2) - dgalpha
    alphahat = (eta_1_nclasses - 1) / etahat_2
    vhat3_cumsum = tf.cumsum(vhat3) - vhat3

    vhat_sum = (tf.concat(((0,), vhat3_cumsum), axis=0) + tf.concat((vhat2, (0,)), axis=0))[
        :, tf.newaxis
    ]

    phi = (
        p_x_neighborhood
        + vhat_sum
        + tf.matmul(lambdahat_2, counts, transpose_b=True)
        - sizefactors * tf.reduce_sum(lambdahat_1, axis=1, keepdims=True)
    )
    pihat = tf.nn.softmax(phi, axis=0)
    pihat_cumsum = tf.cumsum(pihat, axis=0, reverse=True) - pihat

    vhat3_sum = tf.reduce_sum(vhat3)
    gammahat_1 = gamma_1 + pihat @ counts
    gammahat_2 = gamma_2 + tf.matmul(pihat, sizefactors, transpose_b=True)
    etahat_2 = eta_2 - vhat3_sum
    alphahat_1 = 1 + ngenes * tf.reduce_sum(pihat, axis=1)[:-1]
    alphahat_2 = ngenes * tf.reduce_sum(pihat_cumsum, axis=1)[:-1] + alphahat

    elbo = (
        tf.reduce_sum(pihat * p_x_neighborhood)
        + tf.reduce_sum(
            pihat
            * (
                vhat_sum
                + tf.matmul(lambdahat_2, counts, transpose_b=True)
                - tf.reduce_sum(lambdahat_1, axis=1, keepdims=True) * sizefactors
            )
        )
        + tf.reduce_sum((alphahat - alphahat_2) * vhat3)
        + tf.reduce_sum((gamma_1 - gammahat_1) * lambdahat_2)
        + tf.reduce_sum((gammahat_2 - gamma_2) * lambdahat_1)
        - tf.reduce_sum(pihat * phi)
        + tf.reduce_sum(tf.reduce_logsumexp(phi, axis=0))
        - tf.reduce_sum(gammahat_1 * tf.math.log(gammahat_2) - tf.math.lgamma(gammahat_1))
        - tf.reduce_sum((alphahat_1 - 1) * vhat2)
        + tf.reduce_sum(tf.math.lbeta(tf.stack((alphahat_1, alphahat_2), axis=1)))
        - eta_1_nclasses * tf.math.log(etahat_2)
        + (etahat_2 - eta_2) * alphahat
    ) / tf.cast(nclasses * ngenes * tf.shape(counts)[0], counts.dtype)

    return pihat, alphahat_1, alphahat_2, etahat_2, gammahat_1, gammahat_2, elbo


def tissue_segmentation(
    adata: AnnData,
    layer: Optional[str] = None,
    genes: Optional[List[str]] = None,
    sizefactors: Optional[np.ndarray] = None,
    spatial_key: str = "spatial",
    params: TissueSegmentationParameters = TissueSegmentationParameters(),
    labels: Optional[Union[np.ndarray, tf.Tensor]] = None,
    rng: np.random.Generator = np.random.default_rng(),
    copy=False,
) -> Tuple[TissueSegmentation, Union[AnnData, None]]:
    """
    Segment a spatial transcriptomics dataset into distinct spatial regions.

    Uses a hidden Markov random field (HMRF) model with a Poisson likelihood. A Dirichlet process prior allows
    to automatically determine the number of distinct regions in the dataset.

    Args:
        adata: The annotated data matrix.
        layer: Name of the AnnData object layer to use. By default ``adata.X`` is used.
        genes: List of genes to base the segmentation on. Defaults to all genes.
        sizefactors: Scaling factors for the observations. Defaults to total read counts.
        spatial_key: Key in ``adata.obsm`` where the spatial coordinates are stored.
        params: Parameters for the algorithm, e.g. prior distributions, spatial smoothness, etc.
        labels: Initial label for each observation. Defaults to a random initialization.
        rng: Random number generator.
        copy: Whether to return a copy of ``adata`` with results or write the results into ``adata``
            in-place.

    Returns:
        A tuple. The first element is a :py:class:`TissueSegmentation`, the second is ``None`` if ``copy == False``
        or an ``AnnData`` object. Region labels will be in ``adata.obs["segmentation_labels"]`` and region
        probabilities in ``adata.obsm["segmentation_class_probabilities"]``.
    """
    if genes is None and sizefactors is None:
        warnings.warn(
            "Neither genes nor sizefactors are given. Assuming that adata contains complete data set, will calculate size factors and perform segmentation using the complete data set.",
            RuntimeWarning,
        )

    if sizefactors is None:
        sizefactors = calc_sizefactors(adata, layer=layer)
    if genes is not None:
        ngenes = len(genes)
        data = adata[:, genes]
    else:
        ngenes = adata.n_vars
        data = adata
    try:
        X = data.obsm[spatial_key]
    except KeyError:
        X = None

    dtype = tf.float64
    labels_dtype = tf.int32
    nclasses = params.nclasses
    nsamples = data.n_obs
    if nclasses is None:
        nclasses = tf.cast(
            tf.math.ceil(tf.sqrt(tf.convert_to_tensor(nsamples, dtype=tf.float32))), tf.int32
        )
    fngenes = tf.cast(ngenes, dtype=dtype)
    fnclasses = tf.cast(nclasses, dtype=dtype)

    sizefactors = tf.convert_to_tensor(sizefactors[np.newaxis, :], dtype=dtype)

    gamma_1 = tf.convert_to_tensor(params.gamma_1, dtype=dtype)
    gamma_2 = tf.convert_to_tensor(params.gamma_2, dtype=dtype)
    eta_1 = tf.convert_to_tensor(params.eta_1, dtype=dtype)
    eta_2 = tf.convert_to_tensor(params.eta_2, dtype=dtype)

    counts = tf.convert_to_tensor(
        dense_slice(data.X if layer is None else data.layers[layer]), dtype=dtype
    )

    distances = None
    if X is not None and (params.neighbors is None or params.neighbors > 0):
        X = tf.convert_to_tensor(X, dtype=dtype)
        distances = square_distance(X, None)
        if params.neighbors is not None and params.neighbors < nsamples:
            distances, indices = tf.math.top_k(-distances, k=params.neighbors + 1, sorted=True)
            distances = -distances[:, 1:]
            distances = 2 * params.smoothness_factor * tf.reduce_min(distances) / distances
            indices = indices[:, 1:]
            indices = tf.stack(
                (
                    tf.repeat(tf.range(distances.shape[0]), indices.shape[1]),
                    tf.reshape(indices, -1),
                ),
                axis=1,
            )
            dists = tf.reshape(distances, -1)
            distances = tf.scatter_nd(indices, dists, (distances.shape[0], distances.shape[0]))
            distances = tf.tensor_scatter_nd_update(
                distances, indices[:, ::-1], dists
            )  # symmetrize
        else:
            distances = tf.linalg.set_diag(
                distances, tf.repeat(tf.convert_to_tensor(np.inf, dtype), tf.shape(distances)[0])
            )
            distances = 2 * params.smoothness_factor * tf.reduce_min(distances) / distances
    else:
        logging.info("Not using spatial information, fitting Poisson mixture model instead.")

    if labels is not None:
        labels = tf.squeeze(tf.convert_to_tensor(labels, dtype=labels_dtype))
        if tf.rank(labels) > 1 or tf.shape(labels)[0] != nsamples:
            labels = None
            warnings.warn(
                "Shape of given labels does not conform to data. Initializing with random labels.",
                RuntimeWarning,
            )
    if labels is None and distances is not None:
        labels = tf.convert_to_tensor(rng.choice(nclasses, nsamples), dtype=labels_dtype)

    alphahat_1 = tf.ones(shape=(nclasses - 1,), dtype=dtype)
    alphahat_2 = tf.ones(shape=(nclasses - 1,), dtype=dtype)
    etahat_2 = eta_1 + fnclasses - 1
    gammahat_1 = tf.fill((nclasses, ngenes), tf.convert_to_tensor(1e-6, dtype=dtype))
    gammahat_2 = tf.fill((nclasses, 1), tf.convert_to_tensor(1e-6, dtype=dtype))

    prune_threshold = tf.convert_to_tensor(params.class_prune_threshold, dtype=dtype)
    lastelbo = -tf.convert_to_tensor(np.inf, dtype=dtype)
    elbos = []
    nclassestrace = []
    pruneidx = []
    converged = False
    status = TissueSegmentationStatus.MaximumIterationsReached
    for i in range(params.maxiter):
        (pihat, alphahat_1, alphahat_2, etahat_2, gammahat_1, gammahat_2, elbo,) = _segment(
            counts,
            sizefactors,
            distances,
            nclasses,
            fnclasses,
            ngenes,
            labels,
            gamma_1,
            gamma_2,
            eta_1,
            eta_2,
            alphahat_1,
            alphahat_2,
            etahat_2,
            gammahat_1,
            gammahat_2,
        )
        labels = tf.math.argmax(pihat, axis=0, output_type=labels_dtype)
        elbos.append(elbo.numpy())
        nclassestrace.append(nclasses)
        elbodiff = tf.abs(elbo - lastelbo)
        if elbodiff < params.abstol:
            converged = True
            status = TissueSegmentationStatus.AbsoluteToleranceReached
            break
        elif elbodiff / tf.minimum(tf.abs(elbo), tf.abs(lastelbo)) < params.reltol:
            converged = True
            status = TissueSegmentationStatus.AbsoluteToleranceReached
            break
        lastelbo = elbo

        if i == 1 or not i % 10:
            idx, labels = prune_components(labels, pihat, prune_threshold, everything=True)
            if tf.size(idx) < tf.shape(gammahat_1)[0]:
                pruneidx.append(i)
            alphahat_1 = tf.gather(alphahat_1, idx[:-1], axis=0)
            alphahat_2 = tf.gather(alphahat_2, idx[:-1], axis=0)
            gammahat_1 = tf.gather(gammahat_1, idx, axis=0)
            gammahat_2 = tf.gather(gammahat_2, idx, axis=0)
            nclasses = tf.size(idx)

    idx, labels = prune_components(labels, pihat, prune_threshold, everything=True)
    pihat = tf.linalg.normalize(tf.gather(pihat, idx, axis=0), ord=1, axis=0)[0]
    gammahat_1 = tf.gather(gammahat_1, idx, axis=0)
    gammahat_2 = tf.gather(gammahat_2, idx, axis=0)

    if copy:
        adata = adata.copy()
        toreturn = adata
    else:
        toreturn = None
    labels = labels.numpy()
    pihat = pihat.numpy().T
    adata.obs["segmentation_labels"] = pd.Categorical(labels)
    adata.obsm["segmentation_class_probabilities"] = pihat
    return (
        TissueSegmentation(
            converged,
            status,
            labels,
            pihat,
            gammahat_1.numpy(),
            gammahat_2.numpy(),
            i,
            np.asarray(pruneidx),
            np.asarray(elbos),
            np.asarray(nclassestrace),
        ),
        toreturn,
    )



import setuptools
import os
# os.system("SKLEARN_ALLOW_DEPRECATED_SKLEARN_PACKAGE_INSTALL=TRUE")

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="spiderYa",
    version="0.1.6",
    author="Li Shiying",
    author_email="shiyingli7-c@my.cityu.edu.hk",
    description="Identifying spatially variable interactions",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/deepomicslab/SPIDER",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    
   package_data={'spider': ['lrdb/*.tsv'],
                 'spider': ['R_script/*.R']},
    
    install_requires=[
        'anndata>=0.8.0',
        'numpy',
        'Cython',
        'cellrank',
        'fa2',
        'gseapy',
        'h5py',
        'igraph',
        'leidenalg',
        'louvain',
        'matplotlib',
        'matplotlib-venn',
        'networkx',
        'numba',
        'numpy',
        'pandas',
        'plotly',
        'pygco',
        'scanpy',
        'scgco',
        'scikit-learn',
        'scipy',
        'scvelo',
        'seaborn',
        # 'scikit-learn',
        # 'sklearn',
        'somde',
        'somoclu',
        # 'spaotsc',
        'spatialde',
        'stlearn',
        'umap-learn',
        'statsmodels',
        # 'importlib',
        'scprep',
        'squidpy',
        'magic-impute',
        'NaiveDE',
        'gpflow',
        'tensorflow>=2.12',
    ],
)